@@ -17,13 +17,13 @@ def default(val, d):
 
 # relative positional bias
 
-class AlibiPositionalBias(nn.Module):
+class LearnedAlibiPosBias(nn.Module):
     def __init__(self, heads):
         super().__init__()
         self.heads = heads
         slopes = torch.Tensor(self._get_slopes(heads))
         slopes = rearrange(slopes, 'h -> h 1 1')
-        self.register_buffer('slopes', slopes, persistent = False)
+        self.slopes = nn.Parameter(slopes)
         self.register_buffer('bias', None, persistent = False)
 
     def get_bias(self, i, j, device):
@@ -76,61 +76,107 @@ def FeedForward(dim, mult = 4):
         nn.Linear(hidden_dim, dim)
     )
 
+class CausalDepthwiseConv1d(nn.Module):
+    def __init__(self, dim, kernel_size):
+        super().__init__()
+        self.padding = (kernel_size - 1, 0)
+        self.conv = nn.Conv1d(dim, dim, kernel_size = kernel_size, groups = dim)
+
+    def forward(self, x):
+        x = F.pad(x, self.padding)
+        return self.conv(x)
+
 class CausalAttention(nn.Module):
     def __init__(
         self,
         *,
         dim,
         heads = 8,
-        dim_head = 64
+        dim_head = 64,
+        ds_conv_kernel_sizes = (0, 3, 5, 7) # heads were grouped into 4 groups and given a depthwise conv after the queries / keys / values projection
     ):
         super().__init__()
+        assert heads >= 4 and (heads % 4) == 0, 'heads must be greater than 4 and divisible by 4'
+
         self.scale = dim_head ** -0.5
         self.heads = heads
         inner_dim = heads * dim_head
 
         self.norm = nn.LayerNorm(dim)
 
-        self.alibi = AlibiPositionalBias(heads = heads)
+        self.to_qkv = nn.Conv1d(dim, inner_dim * 3, 1, bias = False)
 
-        self.to_q = nn.Sequential(
-            nn.Conv1d(dim, inner_dim, 1, bias = False),
-            nn.Conv1d(inner_dim, inner_dim, 3, bias = False, groups = inner_dim)
-        )
+        # ds convs with different kernel sizes for 4 groups of heads
 
-        self.to_k = nn.Sequential(
-            nn.Conv1d(dim, inner_dim, 1, bias = False),
-            nn.Conv1d(inner_dim, inner_dim, 3, bias = False, groups = inner_dim)
-        )
+        self.groups = len(ds_conv_kernel_sizes)
+        self.qkv_ds_convs = nn.ModuleList([])
 
-        self.to_v = nn.Sequential(
-            nn.Conv1d(dim, inner_dim, 1, bias = False),
-            nn.Conv1d(inner_dim, inner_dim, 3, bias = False, groups = inner_dim)
-        )
+        for _ in range(3): # for queries, keys, values
+            ds_convs = nn.ModuleList([])
+
+            for kernel_size in ds_conv_kernel_sizes:
+                if kernel_size == 0:
+                    ds_convs.append(nn.Identity())
+                    continue
+
+                ds_convs.append(CausalDepthwiseConv1d(inner_dim, kernel_size))
+
+            self.qkv_ds_convs.append(ds_convs)
+
+        # learned alibi positional bias for 4 groups of heads
+
+        self.learned_alibi_pos_biases = nn.ModuleList([LearnedAlibiPosBias(heads = heads // self.groups) for _ in range(self.groups)])
+
+        # outward projection
 
         self.to_out = nn.Linear(inner_dim, dim, bias = False)
 
     def forward(self, x):
         x = self.norm(x)
         x = rearrange(x, 'b n d -> b d n')
-        x = F.pad(x, (2, 0))
 
-        q, k, v = self.to_q(x), self.to_k(x), self.to_v(x)
+        q, k, v = self.to_qkv(x).chunk(3, dim = 1)
+
         q, k, v = rearrange_many((q, k, v), 'b (h d) n -> b h n d', h = self.heads)
 
+        # apply causal depthwise conv to queries, keys, values (a la Primer) with different kernel sizes across 4 groups of heads
+
+        def apply_causal_ds_conv_to_grouped_heads(args):
+            projs, ds_convs = args
+            batch = projs.shape[0]
+
+            projs = rearrange_many(projs.split(self.heads // self.groups, dim = 1), 'b h n d -> (b h) n d')
+            conv_out = [fn(t) for fn, t in zip(ds_convs, projs)]
+            conv_out = map(lambda t: rearrange(t, '(b h) d n -> b h d n', b = batch), conv_out)
+            return torch.cat(tuple(conv_out), dim = 1)
+
+        q, k, v = map(apply_causal_ds_conv_to_grouped_heads, zip((q, k, v), self.qkv_ds_convs))
+
+        # scale and similarity
+
         q = q * self.scale
         sim = einsum('b h i d, b h j d -> b h i j', q, k)
 
-        alibi_bias = self.alibi(sim)
-        sim = sim + alibi_bias
+        # learned alibi pos bias across 4 groups of heads
+        # so heads specialize to looking at different distances of kmers
+
+        grouped_sims = sim.split(self.heads // self.groups, dim = 1)
+        grouped_sims = [(alibi(sim_group) + sim_group) for alibi, sim_group in zip(self.learned_alibi_pos_biases, grouped_sims)]
+        grouped_sims = torch.cat(grouped_sims, dim = 1)
+
+        # causal mask
 
         i, j = sim.shape[-2:]
         causal_mask = torch.ones((i, j), dtype = torch.bool).triu(j - i + 1)
         sim = sim.masked_fill(causal_mask, -torch.finfo(sim.dtype).max)
 
+        # attention, but of course
+
         attn = sim.softmax(dim = -1)
         out = einsum('b h i j, b h j d -> b h i d', attn, v)
 
+        # merge heads
+
         out = rearrange(out, 'b h n d -> b n (h d)')
         return self.to_out(out)
 
@@ -145,7 +191,8 @@ class Tranception(nn.Module):
         num_tokens = 21,
         heads = 8,
         dim_head = 64,
-        ff_mult = 4
+        ff_mult = 4,
+        ds_conv_kernel_sizes = (0, 3, 5, 7)
     ):
         super().__init__()
         self.token_emb = nn.Embedding(num_tokens, dim)
@@ -153,7 +200,7 @@ class Tranception(nn.Module):
         self.layers = nn.ModuleList([])
         for _ in range(depth):
             self.layers.append(nn.ModuleList([
-                CausalAttention(dim = dim, heads = heads, dim_head = dim_head),
+                CausalAttention(dim = dim, heads = heads, dim_head = dim_head, ds_conv_kernel_sizes = ds_conv_kernel_sizes),
                 FeedForward(dim, mult = ff_mult)
             ]))
 
